from wr_attrs import Attr, AttrContainer, _Attrs


def test_can_customise_attrs():
    class MyAttrs(_Attrs):
        pass

    class Base(AttrContainer):
        attrs_cls = MyAttrs

        x = Attr()

    b = Base()
    assert isinstance(b.attrs, MyAttrs)


def test_can_declare_custom_attr_options():
    class MyAttrs(_Attrs):

        @property
        def safe_names(self):
            return [name for name in self.names if self[name].options.get('safe')]

    class Base(AttrContainer):
        attrs_cls = MyAttrs

        x = Attr(safe=True)
        y = Attr(safe=False)
        z = Attr()

    b = Base()
    assert list(b.attrs.names) == ['x', 'y', 'z']
    assert b.attrs.safe_names == ['x']


def test_setter_is_used_to_initialise_attr():
    class A(AttrContainer):
        x = Attr()
        y = Attr()

        @y.setter
        def y(self, value):
            self.attrs.set('y', value * 5)

    a = A(x=3, y=3)
    assert a.x == 3
    assert a.y == 15

    a.attrs.update(x=2, y=2)
    assert a.x == 2
    assert a.y == 10


def test_getter():
    class A(AttrContainer):
        x = Attr()

        @x.getter
        def x(self):
            value = self.attrs.get('x')
            if value is None:
                return None
            else:
                return value * 5

    a = A()
    assert a.x is None

    a.x = 5
    assert a.x == 25
    assert a.x == 25



import pytest

from wr_attrs import Attr, AttrContainer


class A(AttrContainer):
    p = Attr()


class B(A):
    q = Attr()
    r = Attr()


def test_attributes_are_descriptors_for_classes_and_values_for_instances():
    a = A()

    assert isinstance(A.p, Attr)
    assert A.p.name == 'p'
    assert A.p.default is None

    assert a.p is None


def test_attributes_can_be_safely_initialised():
    a = A(p=23)

    assert isinstance(A.p, Attr)
    assert A.p.default is None

    assert a.p == 23


def test_attributes_are_inherited():
    b = B(p=23, q=42)

    assert B.p is A.p

    assert b.p == 23
    assert b.q == 42
    assert b.r is None


def test_attrs_collection():
    b = B(q=42)

    assert len(b.attrs) == 3
    assert list(b.attrs.names) == ['p', 'q', 'r']
    assert list(b.attrs.values) == [('p', None), ('q', 42), ('r', None)]

    assert b.attrs.q == b.attrs['q']
    assert b.attrs.q.name == 'q'
    assert b.attrs.q.default is None
    assert b.attrs.get('q') == 42


def test_default_value_set_in_inherited_class():
    class C(B):
        p = 0
        q = 1

    c = C()

    assert c.p == 0
    assert c.q == 1
    assert c.r is None

    assert C.p is not B.p
    assert C.p.default == 0
    assert B.p.default is None

    assert C.q is not B.q
    assert C.q.default == 1
    assert B.q.default is None

    assert C.r is B.r


def test_custom_setter():
    class D(B):
        s = Attr()

        @s.setter
        def s(self, value):
            self.attrs.set('s', value * 5)

    assert isinstance(D.s, Attr)

    d = D(s=0)
    d.s = 5
    assert d.s == 25


def test_custom_getter():
    class D(B):
        s = Attr()

        @s.getter
        def s(self):
            return self.attrs.get('s') * 10

    d = D(s=5)
    assert d.s == 50

    d.s = 10
    assert d.s == 100


def test_cannot_delete_attr():
    b = B(q=1)
    with pytest.raises(AttributeError) as exc_info:
        delattr(b, 'q')

    assert str(exc_info.value) == '__delete__'


def test_can_inspect_attrs_on_container_class():
    class D(B):
        s = Attr()

    assert D.attrs.owner is D
    assert D.attrs.get('p') is D.p
    assert list(D.attrs.names) == ['p', 'q', 'r', 's']
    assert len(D.attrs) == 4


def test_cannot_set_unknown_attr():
    class D(B):
        pass

    with pytest.raises(AttributeError) as exc_info:
        D(p=1, q=2, r=3, zzz=5)
    assert str(exc_info.value) == 'zzz'

    d = D(p=1, q=2, r=3)

    with pytest.raises(AttributeError) as exc_info:
        d.attrs.set('yyy', 5)
    assert str(exc_info.value) == 'yyy'

    with pytest.raises(AttributeError) as exc_info:
        d.attrs.update(xxx=5)
    assert str(exc_info.value) == 'xxx'


def test_can_set_other_attributes():
    b = B()
    b.something_else = 5
    assert b.something_else == 5



from wr_attrs import Attr, AttrContainer, BoundAttr


class C(AttrContainer):
    x = Attr()

    @x.init
    def x(self, attr: BoundAttr, value=None):
        attr.set(value.upper() if value else 'NOT_SET')


def test_attr_has_value():
    c1 = C()
    assert not c1.attrs.has_value('x')
    c1.x = 'ha'
    assert c1.attrs.has_value('x')

    c2 = C(x='ba')
    assert c2.attrs.has_value('x')
    c2.x = ''
    assert c2.attrs.has_value('x')
    c2.x = None
    assert c2.attrs.has_value('x')

    c3 = C(x=None)
    assert c3.attrs.has_value('x')


def test_attr_init_is_called_on_first_access_only():
    c1 = C()
    assert c1.x == 'NOT_SET'
    c1.x = 'hello'
    assert c1.x == 'hello'

    c2 = C(x='hello')
    assert c2.x == 'HELLO'
    c2.x = 'world'
    assert c2.x == 'world'

    c3 = C()
    c3.x = 'hello'
    assert c3.x == 'HELLO'
    c3.x = 'hello'
    assert c3.x == 'hello'

    c4 = C()
    c4.attrs.update(x='world')
    assert c4.x == 'WORLD'
    c4.x = 'world'
    assert c4.x == 'world'


def test_attr_init_is_called_with_default_value():
    class D(C):
        x = 'hello world'

    assert D.x.default == 'hello world'

    d = D()
    assert not d.attrs.has_value('x')
    assert d.x == 'HELLO WORLD'

    d.x = 'not default'
    assert d.x == 'not default'



from wr_attrs import Attr, AttrContainer, BoundAttr


def test_container_attrs_returns_bound_attr():
    class C(AttrContainer):
        x = Attr()
        y = Attr()

    a = C()
    b = C()

    assert isinstance(a.attrs.x, BoundAttr)
    assert a.attrs.x._owner_ is a

    assert isinstance(b.attrs.x, BoundAttr)
    assert b.attrs.x._owner_ is b

    a.attrs.x.set('hello')
    assert a.x == 'hello'
    assert b.x is None

    b.attrs.x.set('world')
    assert b.x == 'world'
    assert a.x == 'hello'



__version__ = '0.2.0'


import collections
import copy


class Attr:
    """
    Descriptor to declare rich attributes in classes that inherit AttrContainer.

    Deletion of Attr attribute is not permitted because of unclear semantics.
    """

    ALL_ATTRS = '_all_attrs_'

    def __init__(self, name=None, default=None, fget=None, fset=None, finit=None, doc=None, **options):
        self.name = name

        self.default = default

        self.options = options

        self._storage_name = None

        self.fget = fget
        self.fset = fset
        self.finit = finit
        if doc is None and fget is not None:
            doc = fget.__doc__
        self.__doc__ = doc

    @property
    def storage_name(self):
        if self._storage_name is None:
            assert self.name
            self._storage_name = '_{}#{}'.format(self.__class__.__name__, self.name)
        return self._storage_name

    def __get__(self, instance, owner):
        if instance is None:
            return self
        if self.fget is None:
            return instance.attrs.get(self.name, self.default)
        return self.fget(instance)

    def __set__(self, instance, value):
        if self.fset is None:
            instance.attrs.set(self.name, value)
            return
        self.fset(instance, value)

    def __repr__(self):
        return '<{} {!r}>'.format(self.__class__.__name__, self.name)

    def getter(self, fget):
        """
        Decorator to register a getter of Attr.
        The method must have the same name as the Attr.

        In getter, to access the low-level storage, use self.attrs.get(name, default).
        """
        self.fget = fget
        assert self.name is None or fget.__name__ == self.name

        # Must return self so that in class dictionary the getter function doesn't overwrite
        # the descriptor itself.
        return self

    def setter(self, fset):
        """
        Decorator to register a setter of Attr.
        The method must have the same name as the Attr.

        In setter, to access the low-level storage, use self.attrs.set(name, value).
        """
        self.fset = fset
        assert self.name is None or fset.__name__ == self.name

        # Must return self so that in class dictionary the setter function doesn't overwrite
        # the descriptor itself.
        return self

    def init(self, finit):
        """
        Decorator to register an initialiser of Attr.
        The method must have the same name as the Attr.

        Initialiser is called only once for each pair of (AttrContainer instance, Attr).
        It is called the first time attribute value is being set or is being read.
        """

        self.finit = finit
        assert self.name is None or finit.__name__ == self.name

        # Must return self so that in class dictionary the setter function doesn't overwrite
        # the descriptor itself.
        return self

    def __copy__(self):
        return self.__class__(
            name=self.name,
            default=self.default,
            options=dict(self.options),
            fset=self.fset,
            fget=self.fget,
            finit=self.finit,
            doc=self.__doc__,
        )


class BoundAttr:
    """
    An attribute that is bound to an instance of AttrContainer.
    """

    def __init__(self, owner, attr: Attr):
        self._owner_ = owner
        self._attr_ = attr

    def __getattr__(self, name):
        return getattr(self._attr_, name)

    def __setattr__(self, name, value):
        if name in ('_owner_', '_attr_'):
            super().__setattr__(name, value)
        else:
            setattr(self._attr_, name, value)

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def set(self, value):
        return self._owner_.attrs.set(self.name, value)

    def get(self, default=None):
        return self._owner_.attrs.get(self.name, default=default)


class _Attrs:
    """
    A wrapper that simplifies access to the attrs set on container class or instance.
    """

    def __init__(self, owner):
        self.owner = owner

    def set(self, name, value):
        """
        This is the ultimate lowest level method of setting actual attribute value
        in instance dictionary.
        """
        if name not in self.names:
            raise AttributeError(name)
        if isinstance(self.owner, type):
            setattr(self.owner, name, value)
        else:
            if not self.has_value(name) and self[name].finit:
                # Must call initialiser with the value being set.
                # But first must set value to something so that we don't have
                # infinite recursion.
                setattr(self.owner, self[name].storage_name, None)
                self[name].finit(self.owner, self[name], value)
            else:
                setattr(self.owner, self[name].storage_name, value)

    def get(self, name, default=None):
        """
        This is the ultimate lowest level method for getting actual attribute
        value from instance dictionary.
        """
        if isinstance(self.owner, type):
            return getattr(self.owner, name)
        else:
            if not self.has_value(name) and self[name].finit:
                # Must call initialiser with the default value.
                # But first must set value to something so that we don't have
                # infinite recursion.
                setattr(self.owner, self[name].storage_name, None)
                self[name].finit(self.owner, self[name], self[name].default)
            return getattr(self.owner, self[name].storage_name, default)

    def has_value(self, name):
        if isinstance(self.owner, type):
            raise NotImplementedError()
        else:
            return hasattr(self.owner, self[name].storage_name)

    def update(self, *args, **kwargs):
        if args:
            assert len(args) == 1
            assert not kwargs
            kwargs = args[0]
        for key in self.names:
            if key in kwargs:
                setattr(self.owner, key, kwargs.pop(key))
        if kwargs:
            raise AttributeError(list(kwargs.keys())[0])

    @property
    def values(self):
        return ((k, self.get(k)) for k in self.names)

    @property
    def names(self):
        """
        Returns names of all attributes.
        """
        return self.collection.keys()

    @property
    def collection(self):
        return getattr(self.owner, Attr.ALL_ATTRS)

    def __repr__(self):
        return '<{} of {}>'.format(self.__class__.__name__, self.owner)

    def __getattr__(self, name):
        try:
            return BoundAttr(self.owner, self.collection[name])
        except KeyError:
            raise AttributeError(name)

    def __getitem__(self, name):
        return BoundAttr(self.owner, self.collection[name])

    def __contains__(self, name):
        return name in self.collection

    def __len__(self):
        return len(self.collection)

    def __bool__(self):
        return bool(self.collection)


class _AttrContainerMeta(type):
    """
    Meta class for AttrContainer.
    """

    def __new__(meta, name, bases, dct):
        all_attrs = collections.OrderedDict()

        # Populate all_attrs collection with all inherited attributes

        attrs_cls = None

        for base in bases:
            if hasattr(base, 'attrs_cls'):
                attrs_cls = getattr(base, 'attrs_cls')

            if not hasattr(base, Attr.ALL_ATTRS):
                continue

            for attr_name in getattr(base, Attr.ALL_ATTRS):
                attr = getattr(base, attr_name)
                assert isinstance(attr, Attr)
                assert attr.name == attr_name

                if attr_name not in all_attrs:
                    all_attrs[attr_name] = attr

        # Process new class dictionary

        for k in list(dct.keys()):
            if isinstance(dct[k], Attr):
                # Declaration of an attribute.
                # Must set Attr.name if necessary.
                attr = dct[k]
                assert attr.name is None or attr.name == k
                attr.name = k

                # Register Attr in all_attrs collection.
                all_attrs[k] = attr

            elif k in all_attrs:
                # User is setting a class-specific default value for an existing attribute
                new_default = dct[k]
                dct[k] = copy.copy(all_attrs[k])
                dct[k].default = new_default
                all_attrs[k] = dct[k]

        if dct.get('attrs_cls'):
            attrs_cls = dct['attrs_cls']

        dct[Attr.ALL_ATTRS] = all_attrs

        container_cls = super().__new__(meta, name, bases, dct)

        # AttrContainer class also has attrs which are not bounded to any instance.
        container_cls.attrs = attrs_cls(container_cls)  # type: _Attrs

        return container_cls

    @classmethod
    def __prepare__(meta, name, bases):
        # This is only needed for Python 3.5
        return collections.OrderedDict()


class AttrContainer(metaclass=_AttrContainerMeta):
    """
    Attribute container -- base class for user classes that want to use Attr(s).
    """

    attrs_cls = _Attrs

    def __init__(self, **kwargs):
        self.attrs_cls = kwargs.pop('attrs_cls', self.attrs_cls)
        self.attrs = self.attrs_cls(self)  # type: _Attrs
        self.attrs.update(**kwargs)

    def __repr__(self):
        return '<{} at {}>'.format(self.__class__.__name__, id(self))
